-module(vigenere).
-export([encrypt/2, decrypt/2, encrypt/3, decrypt/3, normalize_key/2]).
-include_lib("eunit/include/eunit.hrl").

%%----------------------------------------------------------------------
%% Function: encrypt/3
%% Purpose:  encrypt text using "Vigenère cipher"
%% Args:     Message - text to be encrypted
%%           Key - secret key
%%           Range - tuple with lower and upper bounds of alphabet
%%                   ex. {$A, $Z}
%% Returns:  ecnrypted cipher
%%----------------------------------------------------------------------
encrypt(Message, Key, Range) ->
    NormalizedKey = normalize_key(Key, length(Message)),
    encrypt(Message, NormalizedKey, Range, []).

%%----------------------------------------------------------------------
%% Function: decrypt/3
%% Purpose:  decrypt text using "Vigenère cipher"
%% Args:     Cipher - encrypted text
%%           Key - secret key
%%           Range - tuple with lower and upper bounds of alphabet.
%%                   ex. {$A, $Z}
%% Returns:  decrypted message
%%----------------------------------------------------------------------
decrypt(Cipher, Key, Range) ->
    NormalizedKey = normalize_key(Key, length(Cipher)),
    decrypt(Cipher, NormalizedKey, Range, []).

%%----------------------------------------------------------------------
%% Function: encrypt/2
%% Purpose:  shortcut for encrypt/3. uses A-Z alphabet
%% Args:     Message - text to be encrypted
%%           Key - secret key
%% Returns:  ecnrypted cipher
%%----------------------------------------------------------------------
encrypt(Message, Key) ->
    encrypt(Message, Key, {$A, $Z}).

%%----------------------------------------------------------------------
%% Function: decrypt/2
%% Purpose:  shortcut for decrypt/3. uses A..Z alphabet
%% Args:     Cipher - encrypted text
%%           Key - secret key
%% Returns:  decrypted message
%%----------------------------------------------------------------------
decrypt(Cipher, Key) ->
    decrypt(Cipher, Key, {$A, $Z}).

encrypt([], [], _, Cipher) ->
    lists:reverse(Cipher);
encrypt([M|Message], [K|Key], {L, H}, Cipher) when ((M < L) or (M > H) or (K < L) or (K > H)) ->
    encrypt(Message, Key, {L, H}, [M|Cipher]);
encrypt([M|Message], [K|Key], {L, H}, Cipher) ->
    N = H - L + 1,
    C = (M - L + K - L) rem N + L,
    encrypt(Message, Key, {L, H}, [C|Cipher]).

decrypt([], [], _, Message) ->
    lists:reverse(Message);
decrypt([C|Cipher], [K|Key], {L, H}, Message) when ((C < L) or (C > H) or (K < L) or (K > H)) ->
    decrypt(Cipher, Key, {L, H}, [C|Message]);
decrypt([C|Cipher], [K|Key], {L, H}, Message) ->
    N = H - L + 1,
    M = ((C - K) + N) rem N + L,
    decrypt(Cipher, Key, {L, H}, [M|Message]).

normalize_key(Key, Length) ->
    normalize_key(Key, Key, Length, []).

normalize_key(_, _, 0, Accumulator) ->
    lists:reverse(Accumulator);
normalize_key(Key, [], Length, Accumulator) ->
    normalize_key(Key, Key, Length, Accumulator);
normalize_key(Key, [H|T], Length, Accumulator) ->
    normalize_key(Key, T, Length - 1, [H|Accumulator]).

%%----------------------------------------------------------------------
%% Tests
%%----------------------------------------------------------------------


normalize_key_test_() ->
    [?_assertMatch("PASSWDPAS", normalize_key("PASSWD", 9)),
     ?_assertMatch("PASS WORD", normalize_key("PASS WORD", 9)),
     ?_assertMatch("PASS", normalize_key("PASSWORD", 4))].

encrypt_test_() ->
    [?_assertMatch("LXFOPVEFRNHR", encrypt("ATTACKATDAWN", "LEMON")),
     ?_assertMatch("SPN CHMV SELWW RVX FGCSGY VF NK", encrypt("ALL YOUR BASES ARE BELONG TO US", "SECRET"))].

decrypt_test_() ->
    [?_assertMatch("ATTACKATDAWN", decrypt("LXFOPVEFRNHR", "LEMON")),
     ?_assertMatch("ALL YOUR BASES ARE BELONG TO US", decrypt("SPN CHMV SELWW RVX FGCSGY VF NK", "SECRET"))].
